import click.testing
import pytest
import pathlib
import cards.cli
import shlex
from collections import namedtuple


@pytest.fixture()
def db_empty(tmpdir, monkeypatch):
    fake_home = pathlib.Path(str(tmpdir.mkdir('fake_home')))

    class FakePathLibPath():
        def home(self):
            return fake_home

    monkeypatch.setattr(cards.cli.pathlib, 'Path', FakePathLibPath)


@pytest.fixture()
def cards_cli():
    runner = click.testing.CliRunner()

    def _invoke_cards(input_string):
        input_list = shlex.split(input_string)
        return runner.invoke(cards.cli.cards_cli, input_list).output.rstrip()

    return _invoke_cards


@pytest.fixture()
def db_non_empty(db_empty, cards_cli):
    cards_cli('add "first item"')
    cards_cli('add "second item"')
    cards_cli('add "third item"')


Item = namedtuple('Item', ['id', 'owner', 'priority', 'done', 'summary'])


def items_from_output(output):
    """
    Turn a tabulate output into a tuple of headers and rows
    assuming the output was formatted in the "jira" style
    """
    lines = output.split('\n')
    values = [[item.strip()
               for item in row.split('|')[1:-1]] for row in lines[1:]]
    items = [Item(*v) for v in values]
    return items


@pytest.fixture()
def cards_cli_list_items():
    '''
    Just like cards_cli fixture, with different output.
    Returns detabulated items.
    '''

    runner = click.testing.CliRunner()

    def _invoke_cards(input_string):
        input_list = shlex.split(input_string)
        if 'list' in input_list:
            input_list.append('--format=jira')
        output = runner.invoke(cards.cli.cards_cli, input_list).output.rstrip()
        return items_from_output(output)

    return _invoke_cards



import cards
from cards import Card
import pytest


@pytest.fixture(scope='module', autouse=True)
def db_module(tmp_path_factory):
    '''A db that can be used for all tests'''
    db_dir = tmp_path_factory.mktemp('tmp_db_dir')
    db_location = db_dir / '.cards_db.json'
    cards.set_db_path(db_location)


@pytest.fixture()
def db_empty(db_module):
    cards.delete_all()


def test_add_card(db_empty):
    c = Card(summary='something', owner='okken')
    id = cards.add_card(c)
    assert id is not None
    assert cards.count() == 1


@pytest.fixture()
def db_non_empty(db_empty):
    some_cards = (Card(summary='first item'),
                  Card(summary='second item'),
                  Card(summary='third item'))
    ids = [cards.add_card(c) for c in some_cards]
    return {'initial_cards': some_cards, 'ids': ids}


def test_get_card(db_empty):
    c = Card(summary='something', owner='okken')
    id = cards.add_card(c)

    retrieved_card = cards.get_card(id)

    assert retrieved_card == c
    assert retrieved_card.id == id


def test_list(db_empty):
    some_cards = [Card(summary='one'),
                  Card(summary='two')]
    for c in some_cards:
        cards.add_card(c)

    all_cards = cards.list_cards()
    assert all_cards == some_cards


def test_list_filter_by_priority(db_empty):
    some_cards = [Card(summary='one', priority=2),
                  Card(summary='two'),
                  Card(summary='three', priority=1), ]
    for c in some_cards:
        cards.add_card(c)

    one_and_above = cards.list_cards(filter={'priority': 1})
    two_and_above = cards.list_cards(filter={'priority': 2})
    one, two, three = some_cards
    assert three in one_and_above

    assert one in two_and_above
    assert three in two_and_above
    assert two not in two_and_above


def test_count(db_empty):
    some_cards = [Card(summary='one'),
                  Card(summary='two')]
    for c in some_cards:
        cards.add_card(c)

    assert cards.count() == 2


@pytest.fixture(scope='module')
def four_items(db_module):
    cards.delete_all()
    cards.add_card(Card(summary='one', owner='brian'))
    cards.add_card(Card(summary='two', owner='brian', done=True))
    cards.add_card(Card(summary='three', owner='okken', done=True))
    cards.add_card(Card(summary='three', owner='okken'))
    cards.add_card(Card(summary='four'))


def test_count_no_owner(four_items):
    assert cards.count(noowner=True) == 1


def test_count_owner(four_items):
    assert cards.count(owner='brian') == 2


def test_count_done(four_items):
    assert cards.count(done=True) == 2


def test_count_not_done(four_items):
    assert cards.count(done=False) == 3


def test_update(db_non_empty):
    # GIVEN a card known to be in the db
    all_cards = cards.list_cards()
    a_card = all_cards[0]

    # WHEN we update() the card with new info
    cards.update_card(a_card.id, Card(owner='okken', done=True))

    # THEN we can retrieve the card with get() and
    # and it has all of our changes
    updated_card = cards.get_card(a_card.id)
    expected = Card(summary=a_card.summary, owner='okken', done=True)
    assert updated_card == expected


def test_delete(db_non_empty):
    # GIVEN a non empty db
    a_card = cards.list_cards()[0]
    id = a_card.id
    count_before = cards.count()

    # WHEN we delete one item
    cards.delete_card(id)
    count_after = cards.count()

    # THEN the card is no longer in the db
    all_cards = cards.list_cards()
    assert a_card not in all_cards
    assert count_after == count_before - 1


def test_delete_all(db_non_empty):
    # GIVEN a non empty db

    # WHEN we delete_all()
    cards.delete_all()

    # THEN the count is 0
    count = cards.count()
    assert count == 0



# -*- coding: utf-8 -*-
"""
Tests using cards CLI (command line interface).
This file contains tests related to design changes.
"""

import json
from textwrap import dedent

expected_output_default = """\
  ID  owner    priority    done    summary
----  -------  ----------  ------  -----------
   1                               first item
   2                               second item
   3                               third item"""


def test_list_format(db_non_empty, cards_cli):
    """Check the format of list"""
    output = cards_cli('list')
    print(output)
    assert expected_output_default == output


def test_cards_no_list_arg(db_non_empty, cards_cli):
    """Does `cards` behave just like `cards list`?"""
    output = cards_cli('')
    assert expected_output_default == output


def test_list_json(db_non_empty, cards_cli):
    expected_json = {"cards": [
        {"done": None,
         "id": 1, "owner": None, "priority": None, "summary": "first item"},
        {"done": None,
         "id": 2, "owner": None, "priority": None, "summary": "second item"},
        {"done": None,
         "id": 3, "owner": None, "priority": None, "summary": "third item"}]}
    output_json_str = cards_cli('list --format=json')
    output_json = json.loads(output_json_str)
    assert expected_json == output_json


def test_list_markdown(db_non_empty, cards_cli):
    """Check the format of list --format markdown (and pipe)"""
    expected_output = dedent("""\
    |   ID | owner   | priority   | done   | summary     |
    |-----:|:--------|:-----------|:-------|:------------|
    |    1 |         |            |        | first item  |
    |    2 |         |            |        | second item |
    |    3 |         |            |        | third item  |""")
    output = cards_cli('list -f markdown')
    assert expected_output == output

    output = cards_cli('list -f pipe')
    assert expected_output == output


def test_list_packed(db_non_empty, cards_cli):
    """Check the format of list --format packed"""
    expected_output = dedent("""\
    1 unassigned None o first item
    2 unassigned None o second item
    3 unassigned None o third item""")
    output = cards_cli('list -f packed')
    assert expected_output == output


def test_list_grid(db_non_empty, cards_cli):
    """Check the format of list --format grid"""
    expected_output = dedent("""\
    +------+---------+------------+--------+-------------+
    |   ID | owner   | priority   | done   | summary     |
    +======+=========+============+========+=============+
    |    1 |         |            |        | first item  |
    +------+---------+------------+--------+-------------+
    |    2 |         |            |        | second item |
    +------+---------+------------+--------+-------------+
    |    3 |         |            |        | third item  |
    +------+---------+------------+--------+-------------+""")
    output = cards_cli('list -f grid')
    assert expected_output == output



"""Command Line Interface (CLI) for cards project."""
# todo: refactor: move noowner/owner/done into a filter param for
#       various methods, to get ready for more attributes
# todo: add priority
# todo: add due dates
# todo: add tags
# todo: add history
# todo: list only columns with contents

import os
import click
import cards
import json
import pathlib
from tabulate import tabulate

DEFAULT_TABLEFORMAT = os.environ.get('CARDSTABLEFORMAT', 'simple')


@click.group(invoke_without_command=True,
             context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(version=cards.__version__)
@click.pass_context
def cards_cli(ctx):
    """Run the cards application."""
    if ctx.invoked_subcommand is None:
        ctx.invoke(list_cards, noowner=None, owner=None, done=None)


@cards_cli.command(help="add a card")
@click.argument('summary')
@click.option('-o', '--owner', default=None,
              help='set the card owner')
def add(summary, owner):
    """Add a card to db."""
    set_cards_db_path()
    cards.add_card(cards.Card(summary, owner))


@cards_cli.command(help="delete a card")
@click.argument('card_id', type=int)
def delete(card_id):
    """Remove card in db with given id."""
    set_cards_db_path()
    cards.delete_card(card_id)


@cards_cli.command(name="list", help="list cards")
@click.option('-n', '--noowner', default=None, is_flag=True,
              help='filter on the card without owners')
@click.option('-o', '--owner', default=None,
              help='filter on the card owner')
@click.option('-p', '--priority', default=None,
              type=int,
              help='filter on this priority and above')
@click.option('-d', '--done', default=None,
              type=bool,
              help='filter on cards with given done state')
@click.option('-f', '--format', default=DEFAULT_TABLEFORMAT,
              type=str,
              help='table formatting option, eg. "grid", "simple", "html"')
def list_cards(noowner, owner, priority, done, format):
    """
    List cards in db.
    """
    set_cards_db_path()
    filter = {'noowner': noowner, 'owner': owner,
              'priority': priority, 'done': done}
    the_cards = cards.list_cards(filter=filter)

    #  json is a special case
    if format == 'json':
        items = [c.to_dict() for c in the_cards]
        print(json.dumps({"cards": items}, sort_keys=True, indent=4))
        return

    # who's going to remember 'pipe' for markdown?
    if format == 'markdown':
        format = 'pipe'

    if format == 'packed':
        for t in the_cards:
            done = 'x' if t.done else 'o'
            owner = 'unassigned' if t.owner is None else t.owner
            line = f'{t.id} {owner} {t.priority} {done} {t.summary}'
            print(line)
        return

    # all formats except json/none use tabulate
    items = []
    for t in the_cards:
        done = ' x ' if t.done else ''
        owner = '' if t.owner is None else t.owner
        items.append((t.id, owner, t.priority, done, t.summary))

    print(tabulate(items,
                   headers=('ID', 'owner', 'priority', 'done', 'summary'),
                   tablefmt=format))


@cards_cli.command(help="update card")
@click.argument('card_id', type=int)
@click.option('-o', '--owner', default=None,
              help='change the card owner')
@click.option('-p', '--priority', default=None,
              type=int,
              help='change the card priority')
@click.option('-s', '--summary', default=None,
              help='change the card summary')
@click.option('-d', '--done', default=None,
              type=bool,
              help='change the card done state (True or False)')
def update(card_id, owner, priority, summary, done):
    """Modify a card in db with given id with new info."""
    set_cards_db_path()
    cards.update_card(card_id, cards.Card(summary, owner, priority, done))


@cards_cli.command(help="list count")
@click.option('-n', '--noowner', default=None, is_flag=True,
              help='count cards without owners')
@click.option('-o', '--owner', default=None,
              help='count cards with given owner')
@click.option('-p', '--priority', default=None,
              type=int,
              help='count cards with given priority and above')
@click.option('-d', '--done', default=None,
              type=bool,
              help='count cards with given done state')
def count(noowner, owner, priority, done):
    """Return number of cards in db."""
    set_cards_db_path()
    print(cards.count(noowner, owner, priority, done))


def set_cards_db_path():
    # just put it in users home dir for now
    db_path = pathlib.Path().home() / '.cards_db.json'
    cards.set_db_path(db_path)



# -*- coding: utf-8 -*-
"""
Tests using cards CLI (command line interface).
"""


def test_add(db_empty, cards_cli, cards_cli_list_items):
    # GIVEN an empty database
    # WHEN a new card is added
    cards_cli('add something -o okken')

    # THEN The listing returns just the new card
    items = cards_cli_list_items('list')
    assert len(items) == 1
    assert items[0].summary == 'something'
    assert items[0].owner == 'okken'


def test_list_filter(db_empty, cards_cli, cards_cli_list_items):
    """
    Also kinda tests update
    """
    # GIVEN
    #  two items owned by okken, one that is done
    #  two items with no owner, one that is done
    cards_cli('add -o okken one')
    cards_cli('add -o anyone two')
    cards_cli('add -o okken three')
    cards_cli('add four')
    cards_cli('add five')

    # get the ids for a couple of them
    items = cards_cli_list_items('list')
    for i in items:
        if i.summary in ('three', 'four'):
            cards_cli(f'update {i.id} -d True')

    # `cards --noowner -o okken -d True` should return two items
    items = cards_cli_list_items('list --noowner -o okken -d True')
    assert len(items) == 2
    for i in items:
        assert i.summary in ('three', 'four')
        assert i.done == 'x'
        assert i.owner in ('okken', '')


def test_count(db_empty, cards_cli):
    cards_cli('add one')
    cards_cli('add two')

    assert cards_cli('count') == '2'


def test_delete(db_empty, cards_cli, cards_cli_list_items):
    # GIVEN a db with 2 items
    cards_cli('add one')
    cards_cli('add two')

    an_id = cards_cli_list_items('list')[0].id

    # WHEN we delete one item
    cards_cli(f'delete {an_id}')

    # THEN the other card remains
    assert cards_cli('count') == '1'



import pathlib
import tinydb
from typing import List

from dataclasses import dataclass, field, asdict


__all__ = ["Card", "set_db_path", "get_db_path", "add_card", "get_card",
           "list_cards", "count", "update_card", "delete_card",
           "delete_all"]


@dataclass
class Card:
    summary: str = None
    owner: str = None
    priority: int = None
    done: bool = None
    id: int = field(default=None, compare=False)

    @classmethod
    def from_dict(cls, d):
        return Card(**d)

    def to_dict(self):
        return asdict(self)


_db = None
_db_path = None


def set_db_path(db_path=None):
    global _db
    global _db_path
    if db_path is None:
        _db_path = pathlib.Path().home() / '.cards_db.json'
    else:
        _db_path = db_path
    _db = tinydb.TinyDB(_db_path)


def get_db_path():
    return _db_path


def add_card(card: Card) -> int:
    """Add a card, return the id of card."""
    card.id = _db.insert(card.to_dict())
    _db.update(card.to_dict(), doc_ids=[card.id])
    return card.id


def get_card(card_id: int) -> Card:
    """Return a card with a matching id."""
    return Card.from_dict(_db.get(doc_id=card_id))


def list_cards(filter=None) -> List[Card]:
    """Return a list of all cards."""
    q = tinydb.Query()
    if filter:
        noowner = filter.get('noowner', None)
        owner = filter.get('owner', None)
        priority = filter.get('priority', None)
        done = filter.get('done', None)
    else:
        noowner = None
        owner = None
        priority = None
        done = None
    if noowner and owner:
        results = _db.search(
            (q.owner == owner) |
            (q.owner == None) |  # noqa : "is None" doesn't work for TinyDb
            (q.owner == ''))
    elif noowner or owner == '':
        results = _db.search((q.owner == None) |  (q.owner == '')) # noqa
    elif owner:
        results = _db.search(q.owner == owner)
    elif priority:
        results = _db.search((q.priority != None) &   # noqa
                             (q.priority <= priority))
    else:
        results = _db

    if done is None:
        # return all cards
        return [Card.from_dict(t) for t in results]
    elif done:
        # only done cards
        return [Card.from_dict(t) for t in results if t['done']]
    else:
        # only not done cards
        return [Card.from_dict(t) for t in results if not t['done']]


def count(noowner=None, owner=None, priority=None, done=None) -> int:
    """Return the number of cards in db."""
    filter = {'noowner': noowner, 'owner': owner,
              'priority': priority, 'done': done}
    return len(list_cards(filter=filter))


def update_card(card_id: int, card_mods: Card) -> None:
    """Update a card with modifications."""
    d = card_mods.to_dict()
    changes = {k: v for k, v in d.items() if v is not None}
    _db.update(changes, doc_ids=[card_id])


def delete_card(card_id: int) -> None:
    """Remove a card from db with given card_id."""
    _db.remove(doc_ids=[card_id])


def delete_all() -> None:
    """Remove all tasks from db."""
    _db.purge()



# !/usr/bin/env python
# -*- coding: utf-8 -*-

from io import open
from os import path
import re

from setuptools import find_packages, setup


def read(*parts):
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, *parts), 'r', encoding='utf-8') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="cards",
    version=find_version('src', 'cards', '__init__.py'),
    description='Project task tracking / todo list.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='Brian Okken',
    author_email='brian@pythontesting.net',
    python_requires='>=3.6.0',
    url='https://github.com/okken/pytest-md.git',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    entry_points={
        'console_scripts': ['cards=cards.cli:cards_cli'],
    },
    install_requires=['click',     # for CLI
                      'tabulate',  # for CLI
                      'tinydb',    # for DB
                      "dataclasses; python_version<'3.7'"],
    include_package_data=True,
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
)



import cards
import pathlib


def test_db_path_specified(tmp_path):
    db_location = tmp_path / '.cards_db.json'
    cards.set_db_path(db_location)
    assert cards.get_db_path() == db_location


def test_db_path_default():
    cards.set_db_path()
    expected = pathlib.Path().home() / '.cards_db.json'
    assert cards.get_db_path() == expected
